﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Xml;
using System.Runtime.Serialization.Json;
using System.IO;
using System.Xml.Linq;
using System.Net;
using System.IO.Compression;

namespace CSharpSync2chSample
{
    /// <summary>
    /// このプログラムではお気に入りは構造を持たないので、全ての
    /// お気に入りをリストで持つ。
    /// 構造が制限されるので、リクエストのstructには自分のアプリ名をつけておく。
    /// </summary>
    class Program
    {
        private static string CONFNAME = "\\thread_list.json";

        static void Main(string[] args)
        {
            // 設定の読み込み
            var conf = Sync2chConf.OpenSaveData();

            // 設定再入力
            bool inputID = true;
            if (!string.IsNullOrEmpty(conf.ID))
            {
                Console.WriteLine("現在のID: " + conf.ID);
                Console.Write("[Enter]:OK, [e]:Edit");

                var keyInfo = Console.ReadKey();
                Console.WriteLine();
                if (keyInfo.Key == ConsoleKey.Enter)
                {
                    inputID = false;
                }
            }

            if (inputID)
            {
                Console.Write("IDを入力してください\nID: ");
                conf.ID = Console.ReadLine();
                Console.WriteLine();
            }


            bool inputPass = true;
            if (!string.IsNullOrEmpty(conf.Pass))
            {
                Console.WriteLine("現在のPASS: " + conf.Pass);
                Console.WriteLine("[Enter]:OK, [e]:Edit");

                var keyInfo = Console.ReadKey();
                Console.WriteLine();
                if (keyInfo.Key == ConsoleKey.Enter)
                {
                    inputPass = false;
                }
            }

            if (inputPass)
            {
                Console.Write("API接続用パスワードを入力してください\nPASS: ");
                conf.Pass = Console.ReadLine();
                Console.WriteLine();
            }


            conf.Save();

            if (string.IsNullOrEmpty(conf.ID) || string.IsNullOrEmpty(conf.Pass))
            {
                Console.WriteLine("同期停止");
                return;
            }

            startSync(conf);

            Console.WriteLine("同期終了");
            Console.Out.Flush();

            Console.Read();
        }


        public static void startSync(Sync2chConf conf)
        {

            var thArray = loadThreadInfoArray();
            List<ThreadInfo> list = (thArray == null) ? new List<ThreadInfo>() : new List<ThreadInfo>(thArray);

            var xml = createRequestXml(conf, list);

            Console.WriteLine(xml);

            HttpWebRequest req = (System.Net.HttpWebRequest)
              WebRequest.Create("https://sync2ch.com/api/sync3");
            req.Method = "POST";
            req.ContentType = "text/xml";


            byte[] startData = System.Text.Encoding.Unicode.GetBytes(xml);
            // Console.WriteLine(postData);


            //req.ContentLength = startData.Length;
            //req.Accept = "gzip, deflate";
            req.Headers["Encoding"] = "gzip";
            req.Headers["Accept-Encoding"] = "gzip";
            req.Headers["Authorization"] = "Basic " + Convert.ToBase64String(
                            Encoding.Default.GetBytes(conf.ID + ":" + conf.Pass));
            System.IO.Stream reqStream = req.GetRequestStream();

            reqStream = new GZipStream(reqStream, CompressionMode.Compress);


            reqStream.Write(startData, 0, startData.Length);
            reqStream.Close();

            HttpWebResponse res =
                (System.Net.HttpWebResponse)req.GetResponse();

            if (res.StatusCode != HttpStatusCode.OK)
            {
                throw new Exception("サーバーエラー:" + res.StatusCode);
            }

            Stream resStream = res.GetResponseStream();
            switch (res.ContentEncoding.ToUpperInvariant())
            {
                case "GZIP":
                    resStream = new GZipStream(resStream, CompressionMode.Decompress);
                    break;
            }


            XDocument doc = XDocument.Load(resStream);

            resStream.Close();

            Dictionary<string, XElement> idEntities = new Dictionary<string, XElement>();
            Dictionary<string, ThreadInfo> thInfo = new Dictionary<string, ThreadInfo>();
            list.ForEach((th) => thInfo.Add(th.Url.ToString(), th));

            //レスポンスXML解析
            var rootElement = doc.Root;// doc.Elements().FirstOrDefault();
            if (rootElement != null)
            {
                Dictionary<string, Tuple<string, string>> boardEntities = new Dictionary<string, Tuple<string, string>>();

                foreach (var threadGroup in rootElement.Elements())
                {
                    if (threadGroup.Name == "entities")
                    {

                        var entityelements = threadGroup.Elements().ToArray();
                        foreach (var ele in entityelements)
                        {
                            var element = ele;

                            string id = attr(ele, "id");
                            idEntities[id] = ele;

                        }
                    }
                    else if (threadGroup.Name == "thread_group")
                    {
                        if ("favorite" == attr(threadGroup, "category"))
                        {
                            var newList = new List<ThreadInfo>();

                            //構造をサポートしていないため、子孫スレッドを全て取得し
                            //リストとして処理
                            var entityelements = threadGroup.Descendants().ToArray();
                            foreach (var ele in entityelements)
                            {
                                string id = attr(ele, "id");


                                XElement entityElement = idEntities[id];

                                var status = attr(entityElement, "s");
                                var url = attr(entityElement, "url");
                                var title = attr(entityElement, "title");
                                var now = attr(entityElement, "now");
                                var read = attr(entityElement, "read");
                                var count = attr(entityElement, "count");



                                ThreadInfo th = null;
                                if (!thInfo.ContainsKey(url))
                                {
                                    th = new ThreadInfo();
                                    th.Url = url;
                                }
                                else
                                {
                                    th = thInfo[url];
                                }

                                if (status != "n")
                                {
                                    th.Title = title;
                                    th.CurrentPosition = Int32.Parse(now);
                                    th.ResCount = Int32.Parse(count);
                                    th.ReadPosition = Int32.Parse(read);
                                }
                                newList.Add(th);
                            }

                            list = newList;
                            foreach (var th in list)
                            {
                                Console.WriteLine("title = {0}, url = {1}", th.Title, th.Url);
                            }
                            saveThreadInfoArray(list);
                        }
                    }
                }
            }

            var clientIdAttribute = doc.Root.Attribute("client_id");
            if (clientIdAttribute != null)
            {
                conf.ClientID = clientIdAttribute.Value;
            }

            var syncNumberAttribute = doc.Root.Attribute("sync_number");
            if (syncNumberAttribute != null)
            {
                conf.SyncNumber = syncNumberAttribute.Value;
            }

            conf.Save();// (_saveData);

        }

        public static String attr(XElement el, string name, string def = null)
        {
            if (el == null) return def;

            var tmp = el.Attribute(name);
            if (tmp != null)
            {
                return tmp.Value;
            }
            return def;
        }

        public static void saveThreadInfoArray(List<ThreadInfo> list)
        {
            DataContractJsonSerializer serializer = new DataContractJsonSerializer(typeof(ThreadInfo[]));

            var dirInfo = new DirectoryInfo(Environment.CurrentDirectory);

            FileInfo fileInfo = new FileInfo(dirInfo.FullName + CONFNAME);

            using (Stream stream = (fileInfo != null && fileInfo.Exists) ? fileInfo.Open(FileMode.Truncate, FileAccess.Write) : fileInfo.Create())
            {
                serializer.WriteObject(stream, list.ToArray());
            }
        }

        public static ThreadInfo[] loadThreadInfoArray()
        {

            DataContractJsonSerializer serializer = new DataContractJsonSerializer(typeof(ThreadInfo[]));

            var dirInfo = new DirectoryInfo(Environment.CurrentDirectory);

            FileInfo fileInfo = new FileInfo(dirInfo.FullName + CONFNAME);

            try
            {
                using (Stream stream = (fileInfo != null && fileInfo.Exists) ? fileInfo.OpenRead() : fileInfo.Create())
                {
                    return (ThreadInfo[])serializer.ReadObject(stream);
                }
            }
            catch (System.Runtime.Serialization.SerializationException e)
            {
                Console.WriteLine(e.StackTrace);
            }

            return null;
        }

        public static string createRequestXml(Sync2chConf conf, List<ThreadInfo> thList)
        {
            // XMLの生成
            StringBuilder xmlBuilder = new StringBuilder();
            XmlWriter writer = XmlWriter.Create(xmlBuilder, new XmlWriterSettings() { Indent=true});

       //     writer.WriteProcessingInstruction("xml", "version='1.0' encoding='utf-8'"); //エンコードが指定
            writer.WriteStartElement("sync2ch_request");
            writer.WriteAttributeString("client_id", conf.ClientID);
            writer.WriteAttributeString("sync_number", conf.SyncNumber);
            writer.WriteAttributeString("client_name", "CSharpSample");
            writer.WriteAttributeString("client_version", "1.23");
            writer.WriteAttributeString("os", "Windows");
            //writer.WriteAttributeString("sync_rl", "post");

            // お気に入りの追加
            writer.WriteStartElement("thread_group");
            writer.WriteAttributeString("category", "favorite");
            writer.WriteAttributeString("struct", "CSharpSampleStruct");

            foreach (ThreadInfo thInfo in thList)
            {
                writer.WriteStartElement("th");

                writer.WriteAttributeString("title", thInfo.Title);
                writer.WriteAttributeString("url", thInfo.Url.ToString());
                writer.WriteAttributeString("read", thInfo.ReadPosition + "");
                writer.WriteAttributeString("now", "" + thInfo.CurrentPosition);
                writer.WriteAttributeString("count", "" + thInfo.ResCount);

                writer.WriteEndElement();
            }

            writer.WriteEndElement();


            writer.WriteEndElement();

            writer.Flush();

            return xmlBuilder.ToString();
        }


      
    }
}
